# -*- coding: utf-8 -*-
"""
Renewable Energy Potential(V) EXchange Tool (reVX)
"""
from __future__ import print_function, division, absolute_import
import os

import reVX.plexos as reV_plexos
import reVX.reeds as rev_reeds
import reVX.rpm as rev_rpm

__author__ = """Michael Rossol"""
__email__ = "michael.rossol@nrel.gov"
__version__ = "0.1.1"

REVXDIR = os.path.dirname(os.path.realpath(__file__))
TESTDATADIR = os.path.join(os.path.dirname(REVXDIR), 'tests', 'data')



# -*- coding: utf-8 -*-
"""
Extract ReEDS timeslices from rep-profiles
"""
from copy import deepcopy
from concurrent.futures import as_completed
import json
import logging
import numpy as np
import os
import pandas as pd
from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.utilities import check_tz

from reVX.handlers.outputs import Outputs
from reVX.reeds.reeds_classification import ReedsClassifier
from reVX.reeds.reeds_profiles import ReedsProfiles
from reVX.utilities.exceptions import ReedsValueError, ReedsRuntimeError

logger = logging.getLogger(__name__)


class ReedsTimeslices:
    """
    Create ReEDS timeslices from region-bin-class groups and representative
    profiles
    """
    def __init__(self, profiles, timeslice_map, rev_table=None,
                 reg_cols=('region', 'class')):
        """
        Parameters
        ----------
        profiles : str
            Path to .h5 file containing profiles (representative or cf)
        timeslice_map : str | pandas.DataFrame
            Path to timeslice mapping file or DataFrame with mapping
        rev_table : str | pandas.DataFrame
            reV supply curve or aggregation table, or path to file containing
            table. Only required if timeslice calc is based on all cf_profiles
            instead of representative profiles.
        reg_cols : tuple
            Label(s) for a categorical region column(s) to create timeslices
            for
        """
        self._profiles, self._meta, self._time_index = \
            self._check_profiles(profiles, rev_table=rev_table,
                                 reg_cols=reg_cols)
        self._timeslice_groups = self._map_timeslices(timeslice_map,
                                                      self._time_index)
        if rev_table is not None:
            self._cf_profiles = True
        else:
            self._cf_profiles = False

    def __repr__(self):
        if self._cf_profiles:
            profiles = 'cf profiles'
        else:
            profiles = 'representative profiles'

        msg = ("{} with {} timeslices and {} region-class groups computed "
               "from {}".format(self.__class__.__name__, len(self.timeslices),
                                len(self.regions), profiles))
        return msg

    @property
    def profiles(self):
        """
        .h5 file containing representative profiles

        Returns
        -------
        str
        """
        return self._profiles

    @property
    def meta(self):
        """
        Meta data for "region" groups

        Returns
        -------
        pandas.DataFrame
        """
        return self._meta

    @property
    def regions(self):
        """
        Unique "regions" defaulted to region-class pairs

        Returns
        -------
        list
        """
        return self._meta.index

    @property
    def timeslices(self):
        """
        Timeslice keys

        Returns
        -------
        list
        """
        return sorted(list(self._timeslice_groups.groups.keys()))

    @property
    def profile_datetimes(self):
        """
        Datetime index for representative profiles

        Returns
        -------
        pandas.DatetimeIndex
        """
        return self._time_index

    @staticmethod
    def _extract_meta_and_timeindex(profiles, reg_cols):
        """
        Extract meta data, and time index from the profiles .h5 file

        Parameters
        ----------
        profiles : str
            Path to .h5 file containing profiles (representative or cf)
        reg_cols : tuple
            Label(s) for a categorical region column(s) to create timeslices
            for

        Returns
        -------
        meta : pandas.DataFrame
            Meta data table corresponding to profiles
        time_index : pandas.DatetimeIndex
            Datetime Index for profiles
        """
        with Resource(profiles) as f:
            meta = f.meta
            ignore = ['rep_gen_gid', 'rep_res_gid']
            cols = [c for c in meta.columns if c not in ignore]
            meta = meta[cols]
            isin = [(c in cols) for c in reg_cols]
            if all(isin):
                logger.info('Found region column labels in profile meta for '
                            'timeslice regions: {}'.format(reg_cols))
                meta = meta.set_index(list(reg_cols))

            time_index = f.time_index

        return meta, time_index

    @staticmethod
    def _unpack_list(col):
        """
        Unpack lists stored in given column

        Parameters
        ----------
        col : pandas.DataFrame | pandas.Series
            Column containing lists to unpack

        Returns
        ------
        ndarray
            Vector of values contained in lists
        """
        if isinstance(col.iloc[0], str):
            col = col.apply(json.loads)

        return np.concatenate(col.values)

    @staticmethod
    def _unpack_gen_gids(group_df):
        """
        Unpack gen gids and counts lists

        Parameters
        -----------
        group_df : pandas.DataFrame
            Group DataFrame to unpack gen gids and counts for

        Returns
        -------
        pandas.DataFrame
            DataFrame of unique gen gids and their associated counts
        """
        gen_df = {
            'gid': ReedsTimeslices._unpack_list(group_df['gen_gids']),
            'gid_count': ReedsTimeslices._unpack_list(group_df['gid_counts'])}

        return pd.DataFrame(gen_df).groupby('gid').sum()

    @staticmethod
    def _add_reg_cols(meta, rev_table, reg_cols=('region', 'class')):
        """
        Add reg_cols to meta from rev_table

        Parameters
        ----------
        meta : pandas.DataFrame
            Meta data table corresponding to profiles
        rev_table : None | str | pandas.DataFrame
            reV supply curve or aggregation table,
            or path to file containing table
        reg_cols : tuple
            Label(s) for a categorical region column(s) to create timeslices
            for

        Returns
        -------
        meta : pandas.DataFrame
            Updated meta with reg_cols added to sites to be extract
            timeslices from
        """
        rev_table = ReedsClassifier._parse_table(rev_table)
        rev_table = rev_table.groupby(list(reg_cols))
        rev_table = rev_table.apply(ReedsTimeslices._unpack_gen_gids)

        meta = meta.merge(rev_table.reset_index(), on='gid')
        meta = meta.set_index(list(reg_cols))

        return meta

    @staticmethod
    def _check_profiles(profiles, rev_table=None,
                        reg_cols=('region', 'class')):
        """
        Check profiles to ensure all needed data is available.
        Extract meta and time_index

        Parameters
        ----------
        profiles : str
            Path to .h5 file containing profiles (representative or cf)
        rev_table : None | str | pandas.DataFrame
            reV supply curve or aggregation table,
            or path to file containing table
        reg_cols : tuple
            Label(s) for a categorical region column(s) to create timeslices
            for

        Returns
        -------
        profiles : str
            Path to .h5 file containing profiles (representative or cf)
        meta : pandas.DataFrame
            Meta data table corresponding to profiles
        time_index : pandas.DatetimeIndex
            Datetime Index for rep profiles
        """
        if not profiles.endswith('.h5'):
            msg = ("Cannot parse {} must be a .h5 file!"
                   .format(profiles))
            logger.error(msg)
            raise ReedsValueError(msg)

        if not os.path.isfile(profiles):
            msg = "{} is not a valid file path!".format(profiles)
            logger.error(msg)
            raise ReedsValueError(msg)

        meta, time_index = \
            ReedsTimeslices._extract_meta_and_timeindex(profiles,
                                                        reg_cols)

        if rev_table is not None:
            meta = ReedsTimeslices._add_reg_cols(meta, rev_table,
                                                 reg_cols=reg_cols)
            logger.info('Using input region column labels for timeslice '
                        'regions: {}'.format(reg_cols))

        logger.info('Profile data check complete.')

        return profiles, meta, time_index

    @staticmethod
    def _parse_timeslice_map(timeslice_map):
        """
        Extract timeslice mapping

        Parameters
        ----------
        timeslice_map : str | pandas.DataFrame
            Path to timeslice mapping file or DataFrame with mapping

        Returns
        ----------
        timeslice_map : pandas.GroupBy
            Mapping of each timeslice to a datetime stamp or hour
        """
        if isinstance(timeslice_map, str):
            timeslice_map = pd.read_csv(timeslice_map)
        elif not isinstance(timeslice_map, pd.DataFrame):
            msg = ('Cannot parse timeslice mapping from : {}'
                   .format(type(timeslice_map)))
            logger.error(msg)
            raise ReedsValueError(msg)

        index_col = [c for c in timeslice_map.columns
                     if c in ['datetime', 'hour']]
        if not index_col:
            msg = ('Timeslice mapping must contain a "datetime" or "hour" '
                   'column to enable mapping to profiles datetime index!')
            logger.error(msg)
            raise ReedsRuntimeError(msg)

        if len(index_col) > 1:
            msg = ('Timeslice mapping can only be mapped to "datetime" OR '
                   '"hour"')
            logger.error(msg)
            raise ReedsRuntimeError(msg)

        index = index_col[0]
        timeslice_map = timeslice_map.set_index(index).sort_index()
        if index == 'datetime':
            timeslice_map.index = check_tz(pd.to_datetime(timeslice_map.index))

        return timeslice_map

    @staticmethod
    def _map_timeslices(timeslice_map, time_index):
        """
        Map timeslices to profiles datetime index

        Parameters
        ----------
        timeslice_map : str | pandas.DataFrame
            Path to timeslice mapping file or DataFrame with mapping
        time_index : pandas.DatetimeIndex
            Datetime Index for profiles, must be supplied with rep_profiles
            dictionary

        Returns
        ----------
        timeslice_map : pandas.GroupBy
            Mapping of each timeslice to profiles time_index
        """
        timeslice_map = ReedsTimeslices._parse_timeslice_map(timeslice_map)

        if timeslice_map.index.name == 'datetime':
            mask = timeslice_map.index.isin(time_index)
            if not mask.all():
                msg = ("{} timesteps in the timeslice mapping do not exist "
                       "in the representative profiles datetime index:"
                       "\n{}".format((~mask).sum(),
                                     timeslice_map.index[~mask]))
                logger.error(msg)
                raise ReedsRuntimeError(msg)
        else:
            hour_of_year = ReedsProfiles._get_hour_of_year(time_index)
            hour_max = np.max(hour_of_year)
            timeslice_hours = timeslice_map.index.values
            if np.max(timeslice_hours) > hour_max:
                mask = timeslice_hours > hour_max
                timeslice_hours[mask] = np.abs(hour_max
                                               - timeslice_hours[mask])

            hour_of_year = hour_of_year[np.argsort(hour_of_year)]
            mask = np.searchsorted(hour_of_year, timeslice_hours)
            timeslice_map.index = time_index[mask]

        cols = list(timeslice_map.columns)
        if len(cols) == 1:
            cols = cols[0]

        logger.info('Extracted timeslice map.')

        return timeslice_map.groupby(cols)

    @staticmethod
    def _extract_rep_profiles(profiles_h5, meta):
        """
        Extract representative profiles and combine them with meta data
        and time_index to create a single dataframe to extract stats from

        Parameters
        ----------
        profiles_h5 : str
            Path to .h5 file containing representative profiles
        meta : pandas.DataFrame
            Meta data table for representative profiles

        Returns
        -------
        profiles_df : pandas.DataFrame
            Multi-index DataFrame of profiles for each 'region'
        """
        profiles = {}
        with Resource(profiles_h5) as f:
            time_index = f.time_index
            for ds in f.datasets:
                if 'rep_profiles' in ds:
                    k = int(ds.split('_')[-1])
                    profiles[k] = f[ds]

        cols = [str([int(i) for i in c]) for c in meta.index]
        profiles_df = []
        for k, arr in profiles.items():
            df = pd.DataFrame(arr, columns=cols, index=time_index)
            df.columns = pd.MultiIndex.from_product([[k], df.columns])
            profiles_df.append(df.swaplevel(axis=1))

        profiles_df = pd.concat(profiles_df, axis=1).sort_index(axis=1,
                                                                level=0)

        return profiles_df

    @staticmethod
    def _rep_tslice_stats(ts_profiles):
        """
        Compute means and standard deviations and correlation coefficients
        for each timeslice from representative profiles

        Parameters
        ----------
        ts_profiles : pandas.DataFrame
            Timeseries profiles to compute stats for

        Returns
        -------
        means : pandas.Series
            CF means for each "region"
        stdevs : pandas.Series
            Standard deviations of CF for each "region"
        coeffs : pandas.DataFrame
            Pearson's correlation coefficients between all combinations
            of regions
        """
        cols = sorted(list(set(c[0] for c in ts_profiles.columns)))
        n = len(cols)
        data = ts_profiles.loc[:, (slice(None), 0)][cols].values

        coeffs = np.corrcoef(data, data, rowvar=False)

        coeffs = pd.DataFrame(coeffs[:n, :n], columns=cols, index=cols)
        coeffs = coeffs.fillna(1)

        means = ts_profiles.stack().mean()
        stdevs = ts_profiles.stack().std()

        return means, stdevs, coeffs

    @staticmethod
    def _rep_profile_stats(profiles_h5, meta, timeslice_groups,
                           max_workers=None):
        """
        Compute means and standard divations for each timeslice from
        representative profiles
        Parameters
        ----------
        profiles_h5 : str
            Path to .h5 file containing representative profiles
        meta : pandas.DataFrame
            Meta data table for representative profiles
        timeslice_groups : pandas.GroupBy
            Mapping of each timeslice to profiles time_index
        max_workers : int
            Number of workers to use for parallel computation of means
            and stdevs when using cf profiles.
            1 means run in serial
            None means use all available CPUs

        Returns
        -------
        means : pandas.DataFrame
            Mean CF for each region and timeslice
        stdevs : pandas.DataFrame
            Standard deviation in CF for each region and timeslice
        corr_coeffs : dict
            Correlation matrices for each timeslice
        """

        if max_workers is None:
            max_workers = os.cpu_count()

        profiles = ReedsTimeslices._extract_rep_profiles(profiles_h5, meta)
        means = []
        stdevs = []
        corr_coeffs = {}
        logger.info('Computing representative profile timeslice stats for '
                    '{} timeslice groups.'.format(len(timeslice_groups)))
        logger.info('Computing timeslice stats with max_workers: {}'
                    .format(max_workers))

        if max_workers > 1:
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = {}
                for s, slice_map in timeslice_groups:
                    tslice = profiles.loc[slice_map.index]
                    future = exe.submit(ReedsTimeslices._rep_tslice_stats,
                                        tslice)
                    futures[future] = s

                for i, future in enumerate(as_completed(futures)):
                    s = futures[future]
                    mean, stdev, coeffs = future.result()

                    corr_coeffs[s] = coeffs
                    mean.name = s
                    means.append(mean)

                    stdev.name = s
                    stdevs.append(stdev)

                    logger.info('Completed {} out of {} representative '
                                'profile timeslice stats futures.'
                                .format(i + 1, len(futures)))

        else:
            for i, (s, slice_map) in enumerate(timeslice_groups):
                tslice = profiles.loc[slice_map.index]
                mean, stdev, coeffs = ReedsTimeslices._rep_tslice_stats(tslice)

                corr_coeffs[s] = coeffs
                mean.name = s
                means.append(mean)

                stdev.name = s
                stdevs.append(stdev)

                logger.info('Completed {} out of {} representative '
                            'profile timeslice stats.'
                            .format(i + 1, len(timeslice_groups)))

        means = pd.concat(means, axis=1).T
        means.index.name = 'timeslice'
        stdevs = pd.concat(stdevs, axis=1).T
        stdevs.index.name = 'timeslice'

        return means, stdevs, corr_coeffs

    @staticmethod
    def _cf_group_stats(profiles_h5, region_meta, timeslice_groups):
        """
        Compute means and standard deviations for each region

        Parameters
        ----------
        profiles_h5 : str
            Path to .h5 file containing profiles (representative or cf)
        region_meta : pandas.DataFrame
            Meta data table for subset of cf profiles in specific region
        timeslice_groups : pandas.GroupBy
            Mapping of each timeslice to profiles time_index

        Returns
        -------
        means : pandas.Series
            Mean CF for timeslices in specific region
        stdevs : pandas.Series
            Standard deviation in CF for timeslices in specific region
        """
        gids = region_meta['gid'].values
        counts = region_meta['gid_count'].values
        with Resource(profiles_h5) as f:
            time_index = f.time_index
            profiles = f['cf_profile', :, gids] * counts / np.sum(counts)

        profiles = pd.DataFrame(profiles,
                                index=time_index)
        means = {}
        stdevs = {}
        for s, slice_map in timeslice_groups:
            tslice = profiles.loc[slice_map.index]
            means[s] = tslice.stack().mean()
            stdevs[s] = tslice.stack().std()

        return pd.Series(means), pd.Series(stdevs)

    @staticmethod
    def _cf_profile_stats(profiles_h5, meta, timeslices, max_workers=None):
        """
        Compute timeslice mean and standard deviation from cf profiles

        Parameters
        ----------
        profiles_h5 : str
            Path to .h5 file containing profiles (representative or cf)
        meta : pandas.DataFrame
            Meta data table for cf profiles
        timeslices : pandas.DataFrame
            Mapping between timeslices of interest and profile timesteps

        Returns
        -------
        means : pandas.DataFrame
            Mean CF for each region and timeslice
        stdevs : pandas.DataFrame
            Standard deviation in CF for each region and timeslice
        """
        reg_cols = meta.index.names
        meta = meta.reset_index()

        if max_workers is None:
            max_workers = os.cpu_count()

        logger.info('Computing timeslice stats with max_workers: {}'
                    .format(max_workers))

        if max_workers > 1:
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = {}
                for group, df in meta.groupby(reg_cols):
                    future = exe.submit(ReedsTimeslices._cf_group_stats,
                                        profiles_h5, df, timeslices)
                    futures[future] = group

                logger.debug('Submitted {} futures.'.format(len(futures)))

                means = []
                stdevs = []
                for i, future in enumerate(as_completed(futures)):
                    logger.info('Timeslice future {} out of {} completed.'
                                .format(i + 1, len(futures)))
                    m, s = future.result()
                    group = futures[future]
                    m.name = str(group)
                    s.name = str(group)
                    means.append(m)
                    stdevs.append(s)
        else:
            means = []
            stdevs = []
            for group, df in meta.groupby(reg_cols):
                m, s = ReedsTimeslices._cf_group_stats(profiles_h5, df,
                                                       timeslices)
                m.name = str(list(group))
                s.name = str(list(group))
                means.append(m)
                stdevs.append(s)

        means = pd.concat(means, axis=1)
        means.index.name = 'timeslice'
        stdevs = pd.concat(stdevs, axis=1)
        stdevs.index.name = 'timeslice'

        return means, stdevs

    @staticmethod
    def _flatten_timeslices(table, value_name, reg_cols):
        """
        Flatten timeslice table into legacy ReEDS Format

        Parameters
        ----------
        table : pandas.DataFrame
            Input means or stdev table to flatten
        value_name : str
            Name of values in table
        reg_cols : list
            List of names for "region" classifiers, default is region, class

        Returns
        -------
        out : pandas.DataFrame
            Flattened table
        """
        out = []
        for r, region in table.iteritems():
            if isinstance(r, str):
                r = json.loads(r)

            region.name = value_name
            region = region.to_frame()

            for n, v in zip(reg_cols, r):
                region[n] = v

            out.append(region)

        return pd.concat(out).reset_index()

    @staticmethod
    def _create_correlation_table(corr_coeffs, reg_cols):
        """
        Flatten and combine correlation matrixes for all timeslices

        Parameters
        ----------
        coeffs : dict
            Correlation matrices for each timeslice
        reg_cols : list
            List of names for "region" classifiers, default is region, class

        Returns
        -------
        out : pandas.DataFrame
            Flattened table of correlation coefficients for all timeslices
        """

        logger.info('Creating legacy timeslice correlation table.')
        out = []
        for k, v in corr_coeffs.items():

            ctable = pd.DataFrame({'coefficient': v.values.flatten()})

            second_cols = []
            for i, rc in enumerate(reg_cols):
                r_label_0 = rc
                r_label_1 = rc + '2'
                second_cols.append(r_label_1)

                r0 = [json.loads(c)[i] if isinstance(c, str) else c[i]
                      for c in v.index]
                r1 = [json.loads(c)[i] if isinstance(c, str) else c[i]
                      for c in v.columns]

                r0 = [[r0[i]] * len(r1) for i in range(len(r1))]
                r0 = [item for sublist in r0 for item in sublist]
                r1 *= len(v.index)

                ctable[r_label_0] = r0
                ctable[r_label_1] = r1

            ctable['timeslice'] = k
            corder = reg_cols + second_cols + ['timeslice', 'coefficient']
            ctable = ctable[corder]

            out.append(ctable)

        sort_cols = (reg_cols + ["{}2".format(c) for c in reg_cols]
                     + ['timeslice', ])
        out = pd.concat(out).sort_values(sort_cols).reset_index(drop=True)
        logger.info('Legacy timeslice correlation table complete.')
        return out

    def _stats_to_df(self, means, stdevs):
        """
        Convert means and stdevs to single dataframe

        Parameters
        ----------
        means : pandas.DataFrame
            Mean CF for each region and timeslice
        stdevs : pandas.DataFrame
            Standard deviation in CF for each region and timeslice

        Returns
        -------
        stats : pandas.DataFrame
            Flattened timeslice table with means and sigma
        """
        reg_cols = list(self._meta.index.names)
        means = self._flatten_timeslices(means, 'cfmean', reg_cols)
        stdevs = self._flatten_timeslices(stdevs, 'cfsigma', reg_cols)
        merge_cols = reg_cols + ['timeslice', ]
        stats = means.merge(stdevs, on=merge_cols)
        sort_cols = reg_cols + ['timeslice', ]
        stats = stats.sort_values(sort_cols).reset_index(drop=True)
        return stats

    def _coeffs_to_legacy_format(self, coeffs):
        """
        Convert outputs to legacy format

        Parameters
        ----------
        coeffs : dict
            Correlation matrices for each timeslice, optional

        Returns
        -------
        coeffs : pandas.DataFrame
            Flattened correlation table.
        """
        logger.info('Performing legacy formatting operation.')
        reg_cols = list(self._meta.index.names)

        if coeffs is not None:
            coeffs = self._create_correlation_table(coeffs, reg_cols)

        logger.info('Legacy formatting complete.')

        return coeffs

    @staticmethod
    def _get_correlation_meta(corr, reg_cols):
        """Get the correlation meta data, which is a dataframe of labels for
        the square correlation matrices.

        Parameters
        ----------
        corr : dict
            A dictionary of correlation matrices for each timeslice.
            Each value is a square dataframe with index labels == column labels
        reg_cols : list | tuple
            Labels for the column/index values in the corr dataframe values.
            The order must match the order of the correlation dataframe labels.

        Returns
        -------
        meta : pd.DataFrame
            Meta dataframe with a column for each reg col and a row entry for
            each column/index in the corr dataframe values.
        """

        meta_dict = {c: [] for c in reg_cols}
        corr_labels = corr[list(corr.keys())[0]].columns.values
        for label in corr_labels:
            if isinstance(label, str):
                label = json.loads(label)
            if len(label) != len(reg_cols):
                e = ('Cannot save correlation dataframes. reg_cols was input '
                     'as "{}" but the correlation table has column/index '
                     'values with different label length: {}'
                     .format(reg_cols, corr_labels))
                logger.error(e)
                raise ReedsRuntimeError(e)
            for i, value in enumerate(label):
                meta_dict[reg_cols[i]].append(value)
        meta = pd.DataFrame(meta_dict)
        return meta

    @staticmethod
    def _check_correlation_dfs(corr):
        """Check that all correlation tables have the same index and
        column values.

        Parameters
        ----------
        corr : dict
            A dictionary of correlation matrices for each timeslice.
            Each value is a square dataframe with index labels == column labels
        """
        corr_labels = corr[list(corr.keys())[0]].columns.values
        for df in corr.values():
            if (not all(df.columns.values == corr_labels)
                    or not all(df.index.values == corr_labels)):
                e = 'Correlation matrix labels did not match!'
                logger.error(e)
                raise ReedsRuntimeError(e)
            if df.shape[0] != df.shape[1]:
                e = ('Correlation matrix of shape {} is not square!'
                     .format(df.shape))
                logger.error(e)
                raise ReedsRuntimeError(e)

    @staticmethod
    def sparsify_corr_matrix(arr):
        """Sparsify a symmetric correlation matrix.

        Parameters
        ----------
        arr : np.ndarray
            A 2D symmetric matrix.

        Returns
        -------
        out : np.ndarray
            A 1D matrix representing values from one half of the input arr
            (with the diagnol).
        indices : np.ndarray
            A 1D matrix representing the flat index locations in arr that
            the values in the out array are from.
        """

        if arr.shape[0] != arr.shape[1]:
            e = 'Cannot sparsify a non-square matrix!'
            raise ValueError(e)

        for i in range(len(arr)):
            if arr[i, 0] != arr[0, i]:
                e = 'Matrix does not appear to be symmetric!'
                raise ValueError(e)

        n = arr.shape[0] * arr.shape[1]
        # pylint: disable=no-member
        i_arr = np.triu(np.arange(n).reshape(arr.shape)).flatten()
        indices = i_arr[(i_arr != 0)]
        indices = np.hstack(([0], indices))
        out = arr.flatten()[indices]
        return out, indices

    @staticmethod
    def unsparsify_corr_matrix(arr, indices):
        """UnSparsify a symmetric correlation matrix from a flat matrix
        generated by .sparsify_corr_matrix().

        Parameters
        ----------
        arr : np.ndarray
            A 1D matrix representing values from one half of the input arr
            (with the diagnol).
        indices : np.ndarray
            A 1D matrix representing the flat index locations in arr that
            the values in the out array are from.

        Returns
        -------
        out : np.ndarray
            A 2D symmetric matrix generated from arr.
        """
        if len(arr.shape) > 1:
            e = ('Cannot make a symmetric matrix from a matrix that is '
                 'already 2D!')
            raise ValueError(e)

        indices = indices.astype(int)
        n = int(np.max(indices) + 1)
        a = int(n ** 0.5)
        shape = (a, a)
        out = np.zeros((n, ), dtype=arr.dtype)
        out[indices] = arr
        out = out.reshape(shape)
        for i in range(a):
            for j in range(a):
                out[j, i] = out[i, j]

        return out

    @staticmethod
    def save_correlation_dict(corr, reg_cols, fpath, compression='gzip',
                              sparsify=False):
        """Save a dictionary of correlation coefficient matrices to an h5 file.

        Parameters
        ----------
        corr : dict
            A dictionary of correlation matrices for each timeslice.
            Each value is a square dataframe with index labels == column labels
        reg_cols : list | tuple
            Labels for the column/index values in the corr dataframe values.
            The order must match the order of the correlation dataframe labels.
        fpath : str
            h5 destination filepath (will overwrite).
        compression : str
            H5py dataset compression argument. gzip or None.
        sparsify : bool
            Flag to sparsify symmetric matrices.
        """
        if not isinstance(corr, dict):
            e = ('Tried to save non-dict correlation coefficient data. '
                 'Maybe legacy formatting was still active.')
            logger.error(e)
            raise TypeError(e)

        meta = ReedsTimeslices._get_correlation_meta(corr, reg_cols)
        ReedsTimeslices._check_correlation_dfs(corr)

        with Outputs(fpath, mode='w') as out:
            out['meta'] = meta
            for k, v in corr.items():
                ds_name = 'timeslice_{}'.format(k)
                if compression is not None:
                    data = deepcopy(v.values)
                    data *= 1000
                    data = data.astype(np.int16)
                    if sparsify:
                        data, ind = ReedsTimeslices.sparsify_corr_matrix(data)
                    out.h5.create_dataset(ds_name,
                                          shape=data.shape,
                                          compression=compression,
                                          data=data)
                    out.h5[ds_name].attrs['scale_factor'] = 1000

                    if 'indices' not in out.datasets and sparsify:
                        out.h5.create_dataset('indices',
                                              shape=ind.shape,
                                              compression=compression,
                                              data=ind)
                else:
                    out._create_dset(ds_name,
                                     shape=v.shape,
                                     dtype='int16',
                                     attrs={'scale_factor': 1000})
                    out[ds_name] = v.values

    def compute_stats(self, max_workers=None, legacy_format=True):
        """
        Compute the mean and stdev CF for each timeslice for each "region"
        Compute the correlation coefficients between "regions" from
        timeslice means

        Parameters
        ----------
        max_workers : int
            Number of workers to use for parallel computation of means
            and stdevs when using cf profiles.
            1 means run in serial
            None means use all available CPUs
        legacy_format : bool
            Format outputs into ReEDS legacy format

        Returns
        -------
        stats : pandas.DataFrame
            A single DataFrame containing means and stdevs (sigma)
        corr_coeffs : pandas.DataFrame | dict | NoneType
            In legacy format: a DataFrame of correlation coefficients for
            all pairs of "regions" for each timeslice, else:
            a dictionary of correlation matrices for each timeslice.
            None for CF profiles due to memory constraints
        """
        if self._cf_profiles:
            means, stdevs = self._cf_profile_stats(self._profiles, self._meta,
                                                   self._timeslice_groups,
                                                   max_workers=max_workers)
            coeffs = None
        else:
            means, stdevs, coeffs = \
                self._rep_profile_stats(self._profiles, self._meta,
                                        self._timeslice_groups,
                                        max_workers=max_workers)
        stats = self._stats_to_df(means, stdevs)
        logger.info('Finished timeslice stats computation.')

        if legacy_format:
            coeffs = self._coeffs_to_legacy_format(coeffs)

        return stats, coeffs

    @classmethod
    def run(cls, profiles, timeslice_map, rev_table=None,
            reg_cols=('region', 'class'), max_workers=None,
            legacy_format=False):
        """
        Compute means and standar deviations for each region and timeslice
        from given representative profiles

        Parameters
        ----------
        profiles : str
            Path to .h5 file containing profiles (representative or cf)
        timeslice_map : str | pandas.DataFrame
            Path to timeslice mapping file or DataFrame with mapping
        rev_table : str | pandas.DataFrame
            reV supply curve or aggregation table, or path to file containing
            table. Only required if timeslice calc is based on all cf_profiles
            instead of representative profiles.
        reg_cols : tuple
            Label(s) for a categorical region column(s) to create timeslices
            for
        max_workers : int
            Number of workers to use for parallel computation of means
            and stdevs when using cf profiles.
            1 means run in serial
            None means use all available CPUs
        legacy_format : bool
            Format outputs into ReEDS legacy format

        Returns
        -------
        stats : pandas.DataFrame
            A single DataFrame containing means and stdevs (sigma)
        corr_coeffs : pandas.DataFrame | dict | NoneType
            In legacy format: a DataFrame of correlation coefficients for
            all pairs of "regions" for each timeslice, else:
            a dictionary of correlation matrices for each timeslice.
            None for CF profiles due to memory constraints
        """
        ts = cls(profiles, timeslice_map, rev_table=rev_table,
                 reg_cols=reg_cols)
        stats, coeffs = ts.compute_stats(max_workers=max_workers,
                                         legacy_format=legacy_format)

        return stats, coeffs



"""
Region Classifier Module
"""
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
from shapely.geometry import shape
from scipy.spatial import cKDTree
import logging

from rex import Resource

logger = logging.getLogger(__name__)


class RegionClassifier():
    """
    Base class of region classification

    Examples
    --------
    >>> meta_path = 'meta.csv'
    >>> regions_path = 'us_states.shp'
    >>> regions_label = 'NAME'
    >>>
    >>> classifier = RegionClassifier(meta_path=meta_path,
                                      regions_path=regions_path,
                                      regions_label=regions_label)
    >>>
    >>> force = True
    >>> fout = 'new_meta.csv'
    >>>
    >>> classification = classifier.classify(force=force)
    >>> classifier.output_to_csv(classification, fout)
    """

    CRS = {'init': 'epsg:4326'}
    DEFAULT_REGIONS_LABEL = 'regions_index'

    def __init__(self, meta_path, regions_path, regions_label=None):
        """
        Parameters
        ----------
        meta_path : str | pandas.DataFrame
            Path to meta CSV file, resource .h5 file, or pre-loaded meta
            DataFrame containing lat/lon points
        regions_path : str
            Path to regions shapefile containing labeled geometries
        regions_label : str
            Attribute to use as label in the regions shapefile
        """
        self._regions_label = regions_label

        self._meta = self._get_meta(meta_path)
        self._regions = self._get_regions(regions_path, regions_label)

    @staticmethod
    def output_to_csv(gdf, path):
        """ Export a geopandas dataframe to csv

        Parameters
        ----------
        gdf : GeoPandas DataFrame
            Meta data to export
        path : str
            Output CSV file path for labeled meta CSV file
        """
        output_gdf = gdf.drop('geometry', axis=1)
        if output_gdf.index.name == 'gid':
            output_gdf = output_gdf.reset_index()

        output_gdf.to_csv(path, index=False)

    @staticmethod
    def _get_regions(regions_path, regions_label):
        """ Load the regions shapefile into geopandas dataframe

        Parameters
        ----------
        regions_path : str
            Path to regions shapefile containing labeled geometries
        regions_label : str
            Attribute to use as label in the regions shapefile
        """
        regions = gpd.read_file(regions_path).to_crs(RegionClassifier.CRS)
        if regions_label not in regions.columns:
            regions_label = RegionClassifier.DEFAULT_REGIONS_LABEL
            regions[regions_label] = regions.index
            logger.warning('Setting regions label: {}'.format(regions_label))

        centroids = regions.geometry.centroid
        regions['lon'] = centroids.x
        regions['lat'] = centroids.y

        return regions

    @staticmethod
    def _get_meta(meta_path):
        """ Load the meta csv file into geopandas dataframe

        Parameters
        ----------
        meta_path : str | pandas.DataFrame
            Path to meta CSV file, resource .h5 file, or pre-loaded meta
            DataFrame containing lat/lon points
        """
        if isinstance(meta_path, str):
            if meta_path.endswith('.csv'):
                meta = pd.read_csv(meta_path)
            elif meta_path.endswith('.h5'):
                with Resource(meta_path) as f:
                    meta = f.meta
            else:
                msg = ("Cannot parse meta data from {}, expecting a .csv or "
                       ".h5 file!".format(meta_path))
                logger.error(msg)
                raise RuntimeError(msg)
        elif isinstance(meta_path, pd.DataFrame):
            meta = meta_path
        else:
            msg = ("Cannot parse meta data from {}, expecting a .csv or "
                   ".h5 file path, or a pre-loaded pandas DataFrame"
                   .format(meta_path))
            logger.error(msg)
            raise RuntimeError(msg)

        lat_label, long_label = RegionClassifier._get_lat_lon_labels(meta)
        geometry = [Point(xy) for xy in zip(meta[long_label],
                                            meta[lat_label])]
        meta = gpd.GeoDataFrame(meta, crs=RegionClassifier.CRS,
                                geometry=geometry)

        return meta

    @staticmethod
    def _get_lat_lon_labels(df):
        """ Auto detect the latitude and longitude columns from DataFrame

        Parameters
        ----------
        df : Pandas DataFrame
            Meta data with the latitude/longitude columns to detect
        """

        # Latitude
        lat_col = [c for c in df if c.lower().startswith('lat')]
        if len(lat_col) > 1:
            msg = "Multiple latitude columns found: {}".format(lat_col)
            logger.error(msg)
            raise RuntimeError(msg)

        lat_col = lat_col[0]

        # Longitude
        lon_col = [c for c in df if c.lower().startswith('lon')]
        if len(lon_col) > 1:
            msg = "Multiple longitude columns found: {}".format(lon_col)
            logger.error(msg)
            raise RuntimeError(msg)

        lon_col = lon_col[0]

        return [lat_col, lon_col]

    @staticmethod
    def _nearest(target, lookup):
        """ Lookup the indices to the nearest point

        Parameters
        ----------
        target : Pandas DataFrame
            List of lat/lon points
        lookup : Pandas DataFrame
            List of lat/lon points
        """
        tree = cKDTree(lookup)  # pylint: disable=not-callable
        _, inds = tree.query(target, k=1)

        return inds

    @staticmethod
    def _geom_is_valid(geom):
        """ Check if individual geometry is valid """
        try:
            shape(geom)
            return 1
        except AttributeError:
            return 0

    def classify(self, force=False):
        """ Classify the meta data with regions labels

        Parameters
        ----------
        force : str
            Force outlier classification by finding nearest
        """
        try:
            # Get intersection classifications
            meta_inds, region_inds, outlier_inds = self._intersect()
        except Exception as e:
            logger.warning(e)
            invalid_geom_ids = self._check_geometry()
            if invalid_geom_ids:
                logger.warning('The following geometries are invalid: {}'
                               .format(invalid_geom_ids))
            else:
                logger.exception('Cannot run region classification')
                raise

        # Check for any intersection outliers
        num_outliers = len(outlier_inds)
        if (num_outliers and force):
            # Lookup the nearest region geometry (by centroid)
            logger.warning('The following points are outliers:')
            logger.warning(outlier_inds)
            cols = self._get_lat_lon_labels(self._meta)
            lookup = self._regions[['lat', 'lon']]
            target = self._meta.loc[outlier_inds][cols]
            region_inds += list(self._nearest(target, lookup))

        # Get full list of meta indices and region labels
        meta_inds += outlier_inds
        region_labels = list(self._regions.loc[region_inds,
                                               self._regions_label])
        if (num_outliers and not force):
            # Fill unclassified labels
            region_labels += [-999 for _ in range(num_outliers)]

        # Build classification mapping
        region_labels = np.array(region_labels).astype(str)
        classified_meta = self._meta.loc[meta_inds]
        classified_meta[self._regions_label] = region_labels
        classified_meta.sort_index(inplace=True)

        return classified_meta

    def _intersect(self):
        """ Join the meta points to regions by spatial intersection """

        joined = gpd.sjoin(self._meta, self._regions,
                           how='inner', op='intersects')
        if 'index_left' in joined.columns:
            joined = joined.drop_duplicates('index_left', keep='last')
            meta_inds = list(joined['index_left'])
        else:
            meta_inds = list(joined.index)

        region_inds = list(joined['index_right'])
        outliers = self._meta.loc[~self._meta.index.isin(meta_inds)]
        outlier_inds = list(outliers.index)

        return meta_inds, region_inds, outlier_inds

    def _check_geometry(self):
        """ Get index list of invalid geometries """
        geometry = self._regions.geometry
        isvalid = geometry.apply(self._geom_is_valid)

        return list(self._regions[isvalid == 0].index)

    @classmethod
    def run(cls, meta_path, regions_path, regions_label=None,
            force=False, fout=None):
        """ Run full classification

        Parameters
        ----------
        meta_path : str | pandas.DataFrame
            Path to meta CSV file, resource .h5 file, or pre-loaded meta
            DataFrame containing lat/lon points
        regions_path : str
            Path to regions shapefile containing labeled geometries
        regions_label : str
            Attribute to use a label in the regions shapefile
        force : str
            Force outlier classification by finding nearest
        fout : str
            Output CSV file path for labeled meta CSV file

        Examples
        --------
        >>> meta_path = 'meta.csv'
        >>> regions_path = 'us_states.shp'
        >>> regions_label = 'NAME'
        >>> force = True
        >>> fout = 'new_meta.csv'
        >>>
        >>> RegionClassifier.run(meta_path=meta_path,
                                 regions_path=regions_path,
                                 regions_label=regions_label
                                 force=force, fout=fout)
        """
        classifier = cls(meta_path=meta_path, regions_path=regions_path,
                         regions_label=regions_label)
        classification = classifier.classify(force=force)
        if fout:
            cls.output_to_csv(classification, fout)

        return classification



# -*- coding: utf-8 -*-
"""
Class to handle Supply Curve points
"""
from concurrent.futures import as_completed
import json
import logging
import numpy as np
import os
import pandas as pd
from warnings import warn

from rex.resource import Resource
from rex.utilities import parse_table, SpawnProcessPool

from reVX.utilities.exceptions import SupplyCurvePointCapacityError

logger = logging.getLogger(__name__)


class Point:
    """
    Class to handle single Supply Curve point
    """
    def __init__(self, sc_gid, capacity, res_gids, gid_counts, res_cf):
        """
        Parameters
        ----------
        sc_gid : int
            Supply curve point gid
        capacity : float
            Point capacity
        res_gids : list | str
            Resource gids associated with Supply curve point
        gid_counts : list | str
            Resource gid exclusion pixel counts
        res_cf : ndarray
            Resource gid capacity factor means
        """
        self._sc_gid = int(sc_gid)
        res_order = np.argsort(res_cf)[::-1]
        self._cf_means = res_cf[res_order]
        self._res_gids = self._parse_list(res_gids, dtype=int)[res_order]
        self._gid_counts = self._parse_list(gid_counts, dtype=float)[res_order]
        self._res_capacity = \
            self._gid_counts / np.sum(self._gid_counts) * capacity
        self._avail_cap = self._res_capacity.copy()

    def __repr__(self):
        msg = "{} {}".format(self.__class__.__name__, self.sc_gid)
        return msg

    @property
    def sc_gid(self):
        """
        Supply curve point gid

        Returns
        -------
        int
        """
        return self._sc_gid

    @property
    def capacity(self):
        """
        Supply curve point capacity

        Returns
        -------
        float
        """
        capacity = np.sum(self._avail_cap)
        if np.isclose(capacity, 0):
            capacity = 0.0

        return capacity

    @property
    def resource_gids(self):
        """
        Resource gids associated with supply curve point

        Returns
        -------
        ndarray
        """
        return self._res_gids

    @property
    def gid_counts(self):
        """
        Resource gid exclusion pixel counts

        Returns
        -------
        ndarray
        """
        return self._gid_counts

    @property
    def resource_capacity(self):
        """
        Capacity of resource gids

        Returns
        -------
        ndarray
        """
        return self._res_capacity

    @property
    def available_capacities(self):
        """
        Capacity of available gids

        Returns
        -------
        ndarray
        """
        return self._avail_cap[self._avail_cap > 0]

    @property
    def available_res_gids(self):
        """
        Index of available resource gids

        Returns
        -------
        ndarray
        """
        res_idx = np.where(self._avail_cap > 0)[0]

        return res_idx

    @property
    def cf_means(self):
        """
        Capacity factor means for each resource gid

        Returns
        -------
        ndarray
        """
        return self._cf_means

    @staticmethod
    def _parse_list(list_in, dtype=None):
        """
        Parse json list if needed

        Parameters
        ----------
        list_in : str | list
            List or jsonified list from supply curve table
        dtype : np.dtype, optional
            dtype for output array, if None infer, by default None

        Returns
        -------
        out : ndarary
            vector of values
        """
        if isinstance(list_in, str):
            list_in = json.loads(list_in)
        elif not isinstance(list_in, list):
            msg = ('Must supply either a jsonified list or a list, not {}'
                   .format(type(list_in)))
            logger.error(msg)
            raise ValueError(msg)

        if dtype is None:
            dtype = type(list_in[0])

        out = np.array(list_in, dtype=dtype)

        return out

    def _drop_build_capacity(self, build_capacity, drop=None):
        """
        Drop capacity from Supply Curve point in order of best resource gids
        (based on cf_mean)

        Parameters
        ----------
        build_capacity : float
            Capacity to be built
        drop : int | None, optional
            Number of gids to drop, if None drop all, by default None

        Returns
        -------
        sc_point : pd.Series
            Resource gids being allocated
        capacity : float
            Capacity being allocated
        availability : bool
            Whether Supply Curve point still has available capacity
        """
        s = self.available_res_gids[0]
        if drop is not None:
            e = s + drop
        else:
            e = drop

        drop_slice = slice(s, e, None)
        capacity = self.capacity
        if capacity < build_capacity:
            build_capacity = capacity

        sc_point = {'sc_gid': self.sc_gid,
                    'res_gids': self.resource_gids[drop_slice].tolist(),
                    'gid_counts': self.gid_counts[drop_slice].tolist(),
                    'cf_means': self.cf_means[drop_slice].tolist(),
                    'build_capacity': build_capacity}
        sc_point = pd.Series(sc_point)

        res_caps = self.resource_capacity[drop_slice]
        capacity = np.sum(res_caps)
        if capacity > build_capacity:
            gid_counts = sc_point['gid_counts']
            new_counts = np.round((res_caps[-1] - (capacity - build_capacity))
                                  * np.sum(gid_counts) / capacity)
            gid_counts[-1] = int(new_counts)
            sc_point['gid_counts'] = gid_counts

        self._avail_cap[drop_slice] = 0.0
        availability = self.capacity > 0

        return sc_point, capacity, availability

    def extract_capacity(self, capacity):
        """
        Extract capacity from Supply Curve point in order of best resource gids
        (based on cf_mean)

        Parameters
        ----------
        capacity : float
            Capacity to extract from Supply Curve point

        Returns
        -------
        out : tuple
            (sc_point, capacity, availability)
        """
        build_capacity = capacity.copy()
        if self.capacity > 0:
            if capacity < self.capacity:
                drop = 0
                for cap in self.available_capacities:
                    drop += 1
                    capacity -= cap
                    if capacity <= 0:
                        break
            else:
                drop = None

            out = self._drop_build_capacity(build_capacity, drop=drop)
        else:
            msg = "{} has no remaining capacity".format(self)
            logger.error(msg)
            raise SupplyCurvePointCapacityError(msg)

        return out

    @classmethod
    def create(cls, sc_point, res_cf_means):
        """
        Create Point from supply curve point meta and resource meta

        Parameters
        ----------
        sc_point : pandas.Series
            Supply curve point meta data
        res_cf_means : pandas.Series
            Resource cf_means by gid

        Returns
        -------
        Point
            Initialized Point objects
        """
        sc_gid = sc_point.name
        capacity = sc_point['capacity']
        res_gids = sc_point['res_gids']
        gid_counts = sc_point['gid_counts']
        res_cf = res_cf_means.loc[res_gids].values

        return cls(sc_gid, capacity, res_gids, gid_counts, res_cf)


class SupplyCurvePoints:
    """
    Class to handle Supply Curve points and their meta
    """

    def __init__(self, sc_table, res_meta, max_workers=None,
                 points_per_worker=400, offshore=False):
        """
        Parameters
        ----------
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        res_meta : str | pandas.DataFrame
            Path to resource .h5, generation .h5, or pre-extracted .csv or
            pandas DataFrame
        max_workers : int, optional
            Number of workers to use for point creation, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        """
        self._sc_table = self._parse_sc_table(sc_table, offshore=offshore)

        self._sc_points, self._capacity, self._mask = \
            self._parse_sc_points(self._sc_table, res_meta,
                                  max_workers=max_workers,
                                  points_per_worker=points_per_worker,
                                  offshore=offshore)

    def __repr__(self):
        msg = "{} with {} points".format(self.__class__.__name__, len(self))
        return msg

    def __len__(self):
        return len(self._sc_points)

    def __getitem__(self, sc_gid):
        """
        Extract the Point object for the given sc_gid

        Parameters
        ----------
        sc_gid : int
            Unique Supply Curve Gid

        Returns
        -------
        Point : obj
            Point object for given supply curve gid
        """
        return self._sc_points[sc_gid]

    @property
    def sc_points(self):
        """
        Supply curve points

        Returns
        -------
        dictionary
        """
        return self._sc_points

    @property
    def sc_table(self):
        """
        Supply curve table

        Returns
        -------
        dictionary
        """
        return self._sc_table

    @property
    def sc_gids(self):
        """
        Supply curve point gids

        Returns
        -------
        list
        """
        return list(self.sc_points.keys())

    @property
    def capacity(self):
        """
        Supply curve point capacities

        Returns
        -------
        ndarray
        """
        return self._capacity

    @property
    def mask(self):
        """
        Bool mask of available points

        Returns
        -------
        ndarray
        """
        return self._mask

    @staticmethod
    def _parse_res_meta(res_meta, offshore=False):
        """
        Extract resource meta data from .h5 file or pre-extracted .csv or
        pandas DataFrame

        Parameters
        ----------
        res_meta : str | pandas.DataFrame
            Path to resource .h5, generation .h5, or pre-extracted .csv or
            pandas DataFrame
        offshore : bool, optional
            Include offshore points, by default False

        Returns
        -------
        res_meta : pandas.Series
            Resource gid cf_mean values
        """
        if isinstance(res_meta, str) and res_meta.endswith('.h5'):
            with Resource(res_meta) as f:
                res_meta = f.meta
                res_meta['cf_mean'] = f['cf_mean']
        else:
            res_meta = parse_table(res_meta)
            if 'cf_mean' not in res_meta:
                msg = ("'cf_mean' must be appended to resource meta for "
                       "PLEXOS plant aggregation!")
                logger.error(msg)
                raise RuntimeError(msg)

        if 'offshore' in res_meta:
            if not offshore:
                res_meta = res_meta.loc[res_meta['offshore'] == 0]

        return res_meta.set_index('gid')['cf_mean']

    @staticmethod
    def _parse_sc_table(sc_table, offshore=False):
        """
        Load and clean-up sc_table for use to create Point objects

        Parameters
        ----------
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        offshore : bool, optional
            Include offshore points, by default False

        Returns
        -------
        sc_table : pandas.DataFrame
            Parsed and cleaned supply curve table
        """
        sc_table = parse_table(sc_table)

        if 'offshore' in sc_table:
            if not offshore:
                sc_table = sc_table.loc[sc_table['offshore'] == 0]

        if isinstance(sc_table.iloc[0]['res_gids'], str):
            sc_table.loc[:, 'res_gids'] = \
                sc_table['res_gids'].apply(json.loads).values

        if isinstance(sc_table.iloc[0]['gid_counts'], str):
            sc_table.loc[:, 'gid_counts'] = \
                sc_table['gid_counts'].apply(json.loads).values

        return sc_table

    @staticmethod
    def _create_worker_slices(table, points_per_worker=400):
        """
        Compute the slice of sc_table to submit to each worker

        Parameters
        ----------
        table : pandas.DataFrame
            Table to split across workers
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400

        Returns
        -------
        slices : list
            List of slices of table to submit to each worker
        """
        points = len(table) + 1
        chunks = list(range(0, points, points_per_worker))
        if chunks[-1] < points:
            chunks += [points]

        slices = []
        for s, e in enumerate(chunks[1:]):
            slices.append(slice(chunks[s], e, None))

        return slices

    @staticmethod
    def _create_points(sc_table, res_cf_means, offshore=False,
                       max_workers=None, points_per_worker=400):
        """
        Create Points from all supply curve points in table

        Parameters
        ----------
        sc_table : pandas.DataFrame
            Supply curve table
        res_cf_means : pandas.Series
            Resource cf_means by gid
        offshore : bool, optional
            Include offshore points, by default False
        max_workers : int, optional
            Number of workers to use for point creation, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400

        Returns
        -------
        sc_points : dict
            Dictionary of Points for all supply curve points in sc_table
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        sc_table = SupplyCurvePoints._parse_sc_table(sc_table,
                                                     offshore=offshore)
        if 'sc_gid' in sc_table:
            sc_table = sc_table.set_index('sc_gid')

        cols = ['capacity', 'res_gids', 'gid_counts']
        sc_table = sc_table[cols]

        sc_points = {}
        if max_workers > 1:
            logger.info('Creating supply curve points in parallel')
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                slices = SupplyCurvePoints._create_worker_slices(
                    sc_table, points_per_worker=points_per_worker)
                for sc_slice in slices:
                    table_slice = sc_table.iloc[sc_slice].copy()
                    gids = np.unique(np.hstack(table_slice['res_gids'].values))
                    res_slice = res_cf_means.loc[gids].copy()
                    future = exe.submit(SupplyCurvePoints._create_points,
                                        table_slice,
                                        res_slice,
                                        max_workers=1)
                    futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    sc_points.update(future.result())
                    logger.debug('Completed {} out of {} Points'
                                 .format((i + 1) * points_per_worker,
                                         len(sc_table)))
        else:
            logger.debug('Creating supply curve points in serial')
            for i, (sc_gid, sc_point) in enumerate(sc_table.iterrows()):
                sc_gid = int(sc_gid)
                sc_points[sc_gid] = Point.create(sc_point, res_cf_means)
                logger.debug('Created {} out of {} Points'
                             .format(i + 1, len(sc_table)))

        return sc_points

    @staticmethod
    def _parse_sc_points(sc_table, res_meta, max_workers=None,
                         points_per_worker=400, offshore=False):
        """
        Create a Point instance for all Supply curve points in sc_table.
        Also create vectors of each points capacity and a boolean mask

        Parameters
        ----------
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        res_meta : str | pandas.DataFrame
            Path to resource .h5, generation .h5, or pre-extracted .csv or
            pandas DataFrame
        max_workers : int, optional
            Number of workers to use for point creation, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        offshore : bool, optional
            Include offshore points, by default False

        Returns
        -------
        tuple
            (sc_points, capacity, mask)
        """
        sc_table = SupplyCurvePoints._parse_sc_table(sc_table,
                                                     offshore=offshore)
        if 'sc_gid' in sc_table:
            sc_table = sc_table.set_index('sc_gid')

        res_meta = SupplyCurvePoints._parse_res_meta(res_meta,
                                                     offshore=offshore)
        sc_points = SupplyCurvePoints._create_points(
            sc_table, res_meta,
            offshore=offshore,
            max_workers=max_workers,
            points_per_worker=points_per_worker)

        n = int(sc_table.index.max() + 1)
        capacity = np.zeros(n)
        capacity[sc_table.index] = sc_table['capacity'].values
        mask = np.ones(n, dtype=bool)

        return sc_points, capacity, mask

    def check_sc_gid(self, sc_gid):
        """
        Check to see if sc_gid is still available
        Parameters
        ----------
        sc_gid : int
            Supply curve point gid to check

        Returns
        -------
        bool
            Supply curve point availability
        """
        if not isinstance(sc_gid, int):
            sc_gid = int(sc_gid)

        return self.mask[sc_gid]

    def get_capacity(self, sc_gid, capacity):
        """
        Extract capacity from given Supply Curve point

        Parameters
        ----------
        capacity : float
            Capacity to extract from Supply Curve point

        Returns
        -------
        out : tuple
            sc_point, capacity
        """
        sc_point = self.sc_points[sc_gid]
        try:
            sc_point, capacity, mask = sc_point.extract_capacity(capacity)

            self._mask[sc_gid] = mask
            self._capacity[sc_gid] -= capacity
        except SupplyCurvePointCapacityError as ex:
            logger.warning('WARNING: {}'.format(ex))
            warn(ex)
            capacity = 0.0

        return sc_point, capacity
